# Copyright 2016 NEC Corporation. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_concurrency.fixture import lockutils
from oslo_serialization import jsonutils
from tempest.lib import decorators
from tempest.lib import exceptions

from mistral_tempest_tests.tests import base
from mistral_tempest_tests.tests import utils


class ExecutionTestsV2(base.TestCase):

    _service = 'workflowv2'

    def setUp(self):
        super(ExecutionTestsV2, self).setUp()
        self.useFixture(lockutils.LockFixture('mistral-workflow'))
        _, body = self.client.create_workflow('wf_v2.yaml')

        self.direct_wf_name = 'wf'
        self.direct_wf2_name = 'wf2'
        self.direct_wflong_name = 'wflong'
        self.sub_wf_name = 'subwf1'
        self.direct_wf_id = body['workflows'][0]['id']
        reverse_wfs = [wf for wf in body['workflows'] if wf['name'] == 'wf1']
        self.reverse_wf = reverse_wfs[0]

    def tearDown(self):
        for ex in self.client.executions:
            try:
                self.client.delete_obj('executions', ex, force=True)
            except Exception:
                pass
        self.client.executions = []

        for wf in self.client.workflows:
            try:
                self.client.delete_obj('workflows', wf)
            except Exception:
                pass
        self.client.workflows = []

        super(ExecutionTestsV2, self).tearDown()

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('c0b4b658-6f01-4680-b402-2f683b3d78b6')
    def test_get_list_executions(self):
        resp, body = self.client.get_list_obj('executions')
        self.assertEqual(200, resp.status)
        self.assertNotIn('next', body)

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('0bfcb4b0-b1e4-4499-b81b-0e86c8a2a841')
    def test_get_list_executions_with_pagination(self):
        resp, body = self.client.create_execution(self.direct_wf_name)
        exec_id_1 = body['id']

        self.assertEqual(201, resp.status)

        resp, body = self.client.create_execution(self.direct_wf2_name)
        exec_id_2 = body['id']

        self.assertEqual(201, resp.status)

        resp, body = self.client.get_list_obj('executions')

        self.assertIn(exec_id_1, [ex['id'] for ex in body['executions']])
        self.assertIn(exec_id_2, [ex['id'] for ex in body['executions']])

        resp, body = self.client.get_list_obj(
            'executions?limit=1&sort_keys=workflow_name&sort_dirs=asc')

        self.assertEqual(200, resp.status)
        self.assertEqual(1, len(body['executions']))
        self.assertIn('next', body)

        workflow_name_1 = body['executions'][0].get('workflow_name')
        next_ = body.get('next')
        param_dict = utils.get_dict_from_string(
            next_.split('?')[1],
            delimiter='&'
        )

        # NOTE: 'id' gets included into sort keys automatically with 'desc'
        # sorting to avoid pagination looping.
        expected_dict = {
            'limit': 1,
            'sort_keys': 'workflow_name,id',
            'sort_dirs': 'asc,asc',
        }

        self.assertTrue(
            set(expected_dict.items()).issubset(set(param_dict.items()))
        )

        # Query again using 'next' link
        url_param = next_.split('/')[-1]
        resp, body = self.client.get_list_obj(url_param)

        self.assertEqual(200, resp.status)
        self.assertEqual(1, len(body['executions']))

        workflow_name_2 = body['executions'][0].get('workflow_name')

        self.assertGreater(workflow_name_2, workflow_name_1)

    @decorators.attr(type='sanity')
    @decorators.idempotent_id('5d8ebe04-8de6-414d-908f-213af59e4c6a')
    def test_create_execution_for_direct_wf(self):
        resp, body = self.client.create_execution(self.direct_wf_name)
        exec_id = body['id']
        self.assertEqual(201, resp.status)
        self.assertEqual(self.direct_wf_name, body['workflow_name'])
        self.assertEqual('RUNNING', body['state'])

        resp, body = self.client.get_list_obj('executions')
        self.assertIn(exec_id,
                      [ex_id['id'] for ex_id in body['executions']])

    @decorators.attr(type='sanity')
    @decorators.idempotent_id('101bfdff-8309-4add-9504-544b15f13d95')
    def test_create_execution_for_reverse_wf(self):
        resp, body = self.client.create_execution(
            self.reverse_wf['name'],
            wf_input={self.reverse_wf['input']: "Bye"},
            params={"task_name": "goodbye"})

        exec_id = body['id']
        self.assertEqual(201, resp.status)
        self.assertEqual(self.reverse_wf['name'], body['workflow_name'])
        self.assertEqual('RUNNING', body['state'])

        resp, body = self.client.get_list_obj('executions')
        self.assertIn(exec_id,
                      [ex_id['id'] for ex_id in body['executions']])

        resp, body = self.client.get_object('executions', exec_id)
        # TODO(nmakhotkin): Fix this loop. It is infinite now.
        while body['state'] != 'SUCCESS':
            resp, body = self.client.get_object('executions', exec_id)
            self.assertEqual(200, resp.status)
        self.assertEqual('SUCCESS', body['state'])

    @decorators.attr(type='sanity')
    @decorators.idempotent_id('2df30966-9c45-4a2e-942d-e74bd92cb5aa')
    def test_create_execution_by_wf_id(self):
        resp, body = self.client.create_execution(self.direct_wf_id)
        exec_id = body['id']
        self.assertEqual(201, resp.status)
        self.assertEqual(self.direct_wf_id, body['workflow_id'])
        self.assertEqual('RUNNING', body['state'])

        resp, body = self.client.get_list_obj('executions')
        self.assertIn(
            exec_id,
            [ex_id['id'] for ex_id in body['executions']]
        )

    @decorators.attr(type='sanity')
    @decorators.idempotent_id('f7f50198-2dbd-4ca1-af51-d0eadc1108ac')
    def test_get_execution(self):
        _, execution = self.client.create_execution(self.direct_wf_name)

        resp, body = self.client.get_object('executions', execution['id'])

        del execution['state']
        del body['state']

        self.assertEqual(200, resp.status)
        self.assertEqual(execution['id'], body['id'])

    @decorators.attr(type='sanity')
    @decorators.idempotent_id('2f142ba0-6b88-4d63-8544-05c3dbfe13cc')
    def test_update_execution_pause(self):
        _, execution = self.client.create_execution(self.direct_wflong_name)
        resp, body = self.client.update_execution(
            execution['id'], '{"state": "PAUSED"}')

        self.assertEqual(200, resp.status)
        self.assertEqual('PAUSED', body['state'])

    @decorators.attr(type='sanity')
    @decorators.idempotent_id('f0557236-55ab-457d-9197-05bc2ae53e21')
    def test_update_execution_description(self):
        _, execution = self.client.create_execution(self.direct_wflong_name)
        resp, body = self.client.update_execution(
            execution['id'], '{"description": "description"}')

        self.assertEqual(200, resp.status)
        self.assertEqual('description', body['description'])

    @decorators.attr(type='sanity')
    @decorators.idempotent_id('c54b4d68-b179-4339-bdab-a91cd6e819b7')
    def test_update_execution_fail(self):
        _, execution = self.client.create_execution(self.direct_wflong_name)
        resp, body = self.client.update_execution(
            execution['id'], '{"state": "ERROR", "state_info": "Forced"}')

        self.assertEqual(200, resp.status)
        self.assertEqual('ERROR', body['state'])
        self.assertEqual('Forced', body['state_info'])

    @decorators.attr(type='sanity')
    @decorators.idempotent_id('b5ce0d18-7d78-45bb-813e-ed94cea65fd0')
    def test_update_execution_by_admin(self):
        _, execution = self.client.create_execution(self.direct_wflong_name)
        resp, body = self.admin_client.update_execution(
            execution['id'], '{"description": "description set by admin"}')

        self.assertEqual(200, resp.status)
        self.assertEqual('description set by admin', body['description'])

        resp, body = self.client.get_object('executions', execution['id'])

        self.assertEqual(200, resp.status)
        self.assertEqual("description set by admin", body['description'])

    @decorators.attr(type='sanity')
    @decorators.idempotent_id('c6247362-a082-49ad-a2c3-aaf12419a477')
    def test_update_execution_by_other_fail(self):
        _, execution = self.client.create_execution(self.direct_wf_name)

        self.assertRaises(
            exceptions.NotFound,
            self.alt_client.update_execution,
            execution['id'],
            '{"description": "description set by admin"}'
        )

    @decorators.attr(type='negative')
    @decorators.idempotent_id('d8bde271-6785-4ace-9173-a8a3a01d5eaa')
    def test_get_nonexistent_execution(self):
        self.assertRaises(exceptions.NotFound, self.client.get_object,
                          'executions', '1a2b3c')

    @decorators.attr(type='negative')
    @decorators.idempotent_id('e26e31ba-88cf-4b90-8b3a-fd4ecc612252')
    def test_update_nonexistent_execution(self):
        put_body = '{"state": "STOPPED"}'

        self.assertRaises(exceptions.NotFound,
                          self.client.update_execution,
                          '1a2b3c', put_body)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('b337e270-b3b6-41e2-8de2-05030b06fc37')
    def test_delete_nonexistent_execution(self):
        self.assertRaises(exceptions.NotFound,
                          self.client.delete_obj,
                          'executions', 'nonexist')

    @decorators.attr(type='negative')
    @decorators.idempotent_id('46f7b4b0-7d4a-4bdc-b2b6-46343cdd6f3a')
    def test_create_ex_for_nonexistent_wf(self):
        self.assertRaises(exceptions.NotFound,
                          self.client.create_execution,
                          'nonexist')

    @decorators.attr(type='negative')
    @decorators.idempotent_id('9d27247e-b4d4-40ab-9181-9986655a6be4')
    def test_create_execution_for_reverse_wf_invalid_start_task(self):
        self.assertRaises(
            exceptions.BadRequest,
            self.client.create_execution,
            self.reverse_wf['name'],
            {self.reverse_wf['input']: "Bye"},
            {"task_name": "nonexist"}
        )

    @decorators.attr(type='negative')
    @decorators.idempotent_id('0d6ac42b-4059-40ef-99d0-a65b3cd1837c')
    def test_create_execution_forgot_input_params(self):
        self.assertRaises(
            exceptions.BadRequest,
            self.client.create_execution,
            self.reverse_wf['name'],
            params={"task_name": "nonexist"}
        )

    @decorators.attr(type='sanity')
    @decorators.idempotent_id('52779c73-7563-47b2-8231-a24d6bf531a7')
    def test_action_ex_concurrency(self):
        resp, wf = self.client.create_workflow("wf_action_ex_concurrency.yaml")
        self.assertEqual(201, resp.status)

        wf_name = wf['workflows'][0]['name']
        resp, execution = self.client.create_execution(wf_name)

        self.assertEqual(201, resp.status)
        self.assertEqual('RUNNING', execution['state'])

        self.client.wait_execution_success(execution)

    @decorators.attr(type='sanity')
    @decorators.idempotent_id('eb061c4d-2892-47f0-81e6-37ba15c376bb')
    def test_task_ex_concurrency(self):
        resp, wf = self.client.create_workflow("wf_task_ex_concurrency.yaml")
        self.assertEqual(201, resp.status)

        wf_name = wf['workflows'][0]['name']
        resp, execution = self.client.create_execution(wf_name)

        self.assertEqual(201, resp.status)
        self.assertEqual('RUNNING', execution['state'])

        self.client.wait_execution(execution, target_state='ERROR')

    @decorators.attr(type='sanity')
    @decorators.idempotent_id('acc8e401-2b26-4c41-9e79-8da791da85c0')
    def test_delete_execution_by_admin(self):
        _, body = self.client.create_execution(self.direct_wf_id)
        exec_id = body['id']
        resp, _ = self.admin_client.delete_obj(
            'executions',
            exec_id,
            force=True
        )

        self.assertEqual(204, resp.status)

        self.client.executions.remove(exec_id)

        self.assertRaises(
            exceptions.NotFound,
            self.client.get_object,
            'executions',
            exec_id
        )

    @decorators.idempotent_id('a882876b-7565-4f7f-9714-d99032ffaabb')
    @decorators.attr(type='sanity')
    def test_workflow_execution_of_nested_workflows_within_namespace(self):
        low_wf = 'for_wf_namespace/lowest_level_wf.yaml'
        middle_wf = 'for_wf_namespace/middle_wf.yaml'
        top_wf = 'for_wf_namespace/top_level_wf.yaml'

        resp, wf = self.client.create_workflow(low_wf)
        self.assertEqual(201, resp.status)

        namespace = 'abc'
        resp, wf = self.client.create_workflow(low_wf, namespace=namespace)
        self.assertEqual(201, resp.status)

        resp, wf = self.client.create_workflow(middle_wf)
        self.assertEqual(201, resp.status)

        resp, wf = self.client.create_workflow(top_wf)
        self.assertEqual(201, resp.status)

        resp, wf = self.client.create_workflow(top_wf, namespace=namespace)
        self.assertEqual(201, resp.status)

        wf_name = wf['workflows'][0]['name']
        resp, top_execution = self.client.create_execution(wf_name, namespace)

        self.assertEqual(201, resp.status)
        self.assertEqual('RUNNING', top_execution['state'])
        self.assertEqual(wf_name, top_execution['workflow_name'])
        self.assertEqual(wf_name, top_execution['workflow_name'])
        self.assertEqual(namespace, top_execution['workflow_namespace'])

        self.client.wait_execution(top_execution, target_state='SUCCESS')

        self.assertEqual(
            namespace,
            jsonutils.loads(top_execution['params'])['namespace']
        )

        resp, tasks = self.client.get_tasks(top_execution['id'])
        top_task = tasks['tasks'][0]

        self.assertEqual(wf_name, top_task['workflow_name'])
        self.assertEqual(namespace, top_task['workflow_namespace'])

        resp, executions = self.client.get_executions(top_task['id'])
        middle_execution = executions['executions'][0]

        self.assertEqual('middle_wf', middle_execution['workflow_name'])
        self.assertEqual('', middle_execution['workflow_namespace'])

        self.assertEqual(
            namespace,
            jsonutils.loads(middle_execution['params'])['namespace']
        )

        resp, tasks = self.client.get_tasks(middle_execution['id'])
        middle_task = tasks['tasks'][0]

        self.assertEqual('middle_wf', middle_task['workflow_name'])
        self.assertEqual('', middle_task['workflow_namespace'])

        resp, executions = self.client.get_executions(middle_task['id'])
        lowest_execution = executions['executions'][0]

        self.assertEqual('lowest_level_wf', lowest_execution['workflow_name'])
        self.assertEqual(namespace, lowest_execution['workflow_namespace'])

        self.assertEqual(
            namespace,
            jsonutils.loads(lowest_execution['params'])['namespace']
        )

        resp, tasks = self.client.get_tasks(lowest_execution['id'])
        lowest_task = tasks['tasks'][0]

        self.assertEqual('lowest_level_wf', lowest_task['workflow_name'])
        self.assertEqual(namespace, lowest_task['workflow_namespace'])

        resp, action_executions = self.client.get_action_executions(
            lowest_task['id']
        )

        action_execution = action_executions['action_executions'][0]

        self.assertEqual('lowest_level_wf', action_execution['workflow_name'])
        self.assertEqual(namespace, action_execution['workflow_namespace'])

    @decorators.attr(type='sanity')
    @decorators.idempotent_id('2baa25c5-0b65-4fbe-8d90-2c6599831b6b')
    def test_root_execution_id(self):
        resp, execution = self.client.create_execution(self.sub_wf_name)

        self.assertEqual(201, resp.status)
        self.assertEqual('RUNNING', execution['state'])

        self.client.wait_execution_success(execution)

        resp, body = self.client.get_list_obj(
            'executions?root_execution_id={}'.format(execution['id']))

        self.assertEqual(200, resp.status)
        self.assertNotEmpty(body['executions'])

        self.assertEqual(2, len(body['executions']))

        for exc in body['executions']:
            self.assertEqual(exc['root_execution_id'], execution['id'])



# Copyright 2016 NEC Corporation. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from tempest.lib import decorators
from tempest.lib import exceptions
import testtools

from mistral_tempest_tests.tests import base
from mistral_tempest_tests.tests import utils


class ActionTestsV2(base.TestCase):

    _service = 'workflowv2'

    @staticmethod
    def get_field_value(body, act_name, field):
        return [body['actions'][i][field]
                for i in range(len(body['actions']))
                if body['actions'][i]['name'] == act_name][0]

    def tearDown(self):
        for act in self.client.actions:
            self.client.delete_obj('actions', act)
        self.client.actions = []

        super(ActionTestsV2, self).tearDown()

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('2e1a578a-1c27-409a-96be-84b5c41498cd')
    def test_get_list_actions(self):
        resp, body = self.client.get_list_obj('actions')

        self.assertEqual(200, resp.status)
        self.assertNotEmpty(body['actions'])
        self.assertNotIn('next', body)

    # TODO(rakhmerov): Fix pagination.
    @testtools.skip('Fix pagination of the /actions endpoint')
    @decorators.attr(type='smoke')
    @decorators.idempotent_id('786ee85c-c32d-4ac9-8f45-79ab6bc47ef1')
    def test_get_list_actions_with_pagination(self):
        resp, body = self.client.get_list_obj(
            'actions?limit=1&sort_keys=name&sort_dirs=desc'
        )

        self.assertEqual(200, resp.status)
        self.assertEqual(1, len(body['actions']))
        self.assertIn('next', body)

        name_1 = body['actions'][0].get('name')
        next_ = body.get('next')

        param_dict = utils.get_dict_from_string(
            next_.split('?')[1],
            delimiter='&'
        )

        # NOTE: 'id' gets included into sort keys automatically with 'desc'
        # sorting to avoid pagination looping.
        expected_sub_dict = {
            'limit': 1,
            'sort_keys': 'name,id',
            'sort_dirs': 'desc,asc'
        }

        self.assertEqual(param_dict, param_dict | expected_sub_dict)

        # Query again using 'next' hint
        url_param = next_.split('/')[-1]
        resp, body = self.client.get_list_obj(url_param)

        self.assertEqual(200, resp.status)
        self.assertEqual(1, len(body['actions']))

        name_2 = body['actions'][0].get('name')

        self.assertGreater(name_1, name_2)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('5148358e-200f-49c7-8e88-1ddeec61c6a9')
    def test_get_list_actions_nonexist_sort_dirs(self):
        context = self.assertRaises(
            exceptions.BadRequest,
            self.client.get_list_obj,
            'actions?limit=1&sort_keys=id&sort_dirs=nonexist'
        )

        self.assertIn(
            'Unknown sort direction',
            context.resp_body.get('faultstring')
        )

    @decorators.attr(type='negative')
    @decorators.idempotent_id('85482ce8-70f4-47a6-9e80-de1ac22b6412')
    def test_get_list_actions_invalid_limit(self):
        context = self.assertRaises(
            exceptions.BadRequest,
            self.client.get_list_obj,
            'actions?limit=-1&sort_keys=id&sort_dirs=asc'
        )

        self.assertIn(
            'Limit must be positive',
            context.resp_body.get('faultstring')
        )

    @decorators.attr(type='negative')
    @decorators.idempotent_id('a203e75b-2013-422f-b9eb-da4375041058')
    def test_get_list_actions_duplicate_sort_keys(self):
        context = self.assertRaises(
            exceptions.BadRequest,
            self.client.get_list_obj,
            'actions?limit=1&sort_keys=id,id&sort_dirs=asc,asc'
        )

        self.assertIn(
            'Length of sort_keys must be equal or greater than sort_dirs',
            context.resp_body.get('faultstring')
        )

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('9a53af71-8f1e-4ad5-b572-2c4c621715c0')
    def test_get_list_actions_equal_to_filter(self):
        resp, body = self.client.create_action('action_v2.yaml')

        self.assertEqual(201, resp.status)

        resp, body = self.client.get_list_obj('actions?name=greeting')

        self.assertEqual(200, resp.status)
        self.assertNotEmpty(body['actions'])
        self.assertEqual(1, len(body['actions']))

        self.assertEqual('greeting', body['actions'][0]['name'])

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('3c3d28ce-9490-41ae-a918-c28f843841e1')
    def test_get_list_actions_not_equal_to_filter(self):
        resp, body = self.client.create_action('action_v2.yaml')

        self.assertEqual(201, resp.status)

        resp, body = self.client.get_list_obj('actions?name=neq:greeting')

        self.assertEqual(200, resp.status)
        self.assertNotEmpty(body['actions'])
        self.assertGreater(len(body['actions']), 0)

        for act in body['actions']:
            self.assertNotEqual('greeting', act['name'])

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('84823a84-5caa-427d-8a2c-622a1d1893b1')
    def test_get_list_actions_in_list_filter(self):
        resp, body = self.client.create_action('action_v2.yaml')

        self.assertEqual(201, resp.status)

        resp, body = self.client.get_list_obj(
            'actions?name=in:greeting,std.echo'
        )

        self.assertEqual(200, resp.status)
        self.assertEqual(2, len(body['actions']))

        for act in body['actions']:
            self.assertIn(act['name'], ['greeting', 'std.echo'])

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('4b05dfcf-ef39-4032-9528-c8422c7329dd')
    def test_get_list_actions_not_in_list_filter(self):
        resp, body = self.client.create_action('action_v2.yaml')

        self.assertEqual(201, resp.status)

        resp, body = self.client.get_list_obj(
            'actions?name=nin:greeting,std.echo'
        )

        self.assertEqual(200, resp.status)
        self.assertGreater(len(body['actions']), 0)

        for act in body['actions']:
            self.assertNotIn(act['name'], ['greeting', 'std.echo'])

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('20b3d527-447d-492b-8cb7-ac5e3757d7d5')
    def test_get_list_actions_greater_than_filter(self):
        resp, body = self.client.create_action('action_v2.yaml')

        self.assertEqual(201, resp.status)

        # NOTE(rakhmerov): The filter "name=gt:freeting" might seem weird
        # but we can use it because Mistral compare alphabetical order
        # of action names with filter value.
        resp, body = self.client.get_list_obj('actions?name=gt:freeting')

        self.assertEqual(200, resp.status)

        # There should be at least "greeting" action in the result set
        # and possible some other standard actions whose names go after
        # "freeting" in the alphabet order.
        self.assertGreater(len(body['actions']), 0)

        # Make sure the "greeting" action is in the result set.
        self.assertEqual(
            1,
            len([act for act in body['actions'] if act['name'] == 'greeting'])
        )

    # TODO(rakhmerov): Come up with a good test, we can't use 'created_at'
    # anymore.
    @testtools.skip('Come up with a good test')
    @decorators.attr(type='smoke')
    @decorators.idempotent_id('7f598dba-f169-47ec-a487-f0ed31484aff')
    def test_get_list_actions_greater_than_equal_to_filter(self):
        resp, body = self.client.create_action('action_v2.yaml')

        self.assertEqual(201, resp.status)

        created_acts = [action['name'] for action in body['actions']]

        _, body = self.client.get_object('actions', created_acts[0])

        time = body['created_at']

        resp, body = self.client.get_list_obj(
            'actions?created_at=gte:' + time.replace(' ', '%20')
        )

        actions = [action['name'] for action in body['actions']]

        self.assertEqual(200, resp.status)
        self.assertIn(created_acts[0], actions)

    # TODO(rakhmerov): Come up with a good test, we can't use 'created_at'
    # anymore.
    @testtools.skip('Come up with a good test')
    @decorators.attr(type='smoke')
    @decorators.idempotent_id('874fb57d-a762-4dc3-841d-396657510d23')
    def test_get_list_actions_less_than_filter(self):
        resp, body = self.client.create_action('action_v2.yaml')

        self.assertEqual(201, resp.status)

        created_acts = [action['name'] for action in body['actions']]

        _, body = self.client.get_object('actions', created_acts[0])

        time = body['created_at']

        resp, body = self.client.get_list_obj(
            'actions?created_at=lt:' + time.replace(' ', '%20')
        )

        actions = [action['name'] for action in body['actions']]

        self.assertEqual(200, resp.status)
        self.assertNotIn(created_acts[0], actions)

    # TODO(rakhmerov): Come up with a good test, we can't use 'created_at'
    # anymore.
    @testtools.skip('Come up with a good test')
    @decorators.attr(type='smoke')
    @decorators.idempotent_id('1fda6c31-b0c3-4b78-9f67-b920e1f6c973')
    def test_get_list_actions_less_than_equal_to_filter(self):
        resp, body = self.client.create_action('action_v2.yaml')

        self.assertEqual(201, resp.status)

        created_acts = [action['name'] for action in body['actions']]

        _, body = self.client.get_object('actions', created_acts[0])

        time = body['created_at']

        resp, body = self.client.get_list_obj(
            'actions?created_at=lte:' + time.replace(' ', '%20')
        )

        actions = [action['name'] for action in body['actions']]

        self.assertEqual(200, resp.status)
        self.assertIn(created_acts[0], actions)

    @decorators.attr(type='smoke')
    @decorators.idempotent_id('cbb716f1-7fc7-4884-8fa9-6ff2bc35ee29')
    def test_get_list_actions_multiple_filter(self):
        resp, body = self.client.create_action('action_v2.yaml')

        self.assertEqual(201, resp.status)

        # TODO(rakhmerov): Both filters are temporarily related to the
        # field 'name', there's no other suitable field right now to
        # to apply a filter to. We could use "&tags=has:hello", for
        # example, but it will work only after the Action Provider
        # transition is completed, now this filter is broken.
        resp, body = self.client.get_list_obj(
            'actions?name=in:greeting,farewell&name=greeting'
        )

        self.assertEqual(200, resp.status)
        self.assertEqual(1, len(body['actions']))
        self.assertEqual('greeting', body['actions'][0]['name'])

    @decorators.attr(type='negative')
    @decorators.idempotent_id('45fdc1f3-4d89-4035-9b76-08ef94c92628')
    def test_get_list_actions_invalid_filter(self):
        self.assertRaises(
            exceptions.BadRequest,
            self.client.get_list_obj,
            'actions?is_system<False'
        )
        self.assertRaises(
            exceptions.BadRequest,
            self.client.get_list_obj,
            'actions?is_system!=False'
        )
        self.assertRaises(
            exceptions.BadRequest,
            self.client.get_list_obj,
            'actions?created_at>2016-02-23%2008:51:26'
        )

    @decorators.attr(type='sanity')
    @decorators.idempotent_id('5dbceaf3-6a32-4a4f-9427-1bbdb6f3c574')
    def test_create_and_delete_few_actions(self):
        resp, body = self.client.create_action('action_v2.yaml')
        self.assertEqual(201, resp.status)

        created_acts = [action['name'] for action in body['actions']]

        resp, body = self.client.get_list_obj('actions')
        self.assertEqual(200, resp.status)

        actions = [action['name'] for action in body['actions']]

        for act in created_acts:
            self.assertIn(act, actions)
            self.client.delete_obj('actions', act)

        _, body = self.client.get_list_obj('actions')
        actions = [action['name'] for action in body['actions']]

        for act in created_acts:
            self.assertNotIn(act, actions)
            self.client.actions.remove(act)

    @decorators.attr(type='sanity')
    @decorators.idempotent_id('d7dad5de-6b1f-4813-b789-78f075252639')
    def test_get_action(self):
        _, body = self.client.create_action('action_v2.yaml')
        action_name = body['actions'][0]['name']
        resp, body = self.client.get_object('actions', action_name)

        self.assertEqual(200, resp.status)
        self.assertEqual(action_name, body['name'])

    @decorators.attr(type='sanity')
    @decorators.idempotent_id('21a031c8-8e2d-421f-8dfe-71a3b5e44381')
    def test_update_action(self):
        _, body = self.client.create_action('action_v2.yaml')
        action = body['actions'][0]['name']

        act_created_at = self.get_field_value(
            body=body, act_name=action, field='created_at')

        self.assertNotIn('updated at', body['actions'])

        resp, body = self.client.update_request('actions', 'action_v2.yaml')
        self.assertEqual(200, resp.status)

        actions = [act['name'] for act in body['actions']]
        self.assertIn(action, actions)

        updated_act_created_at = self.get_field_value(
            body=body, act_name=action, field='created_at')

        self.assertEqual(act_created_at.split(".")[0], updated_act_created_at)
        self.assertTrue(all(['updated_at' in item
                             for item in body['actions']]))

    @decorators.attr(type='sanity')
    @decorators.idempotent_id('329b1030-c55c-45f0-8129-cc892bc23dcc')
    def test_get_action_definition(self):
        _, body = self.client.create_action('action_v2.yaml')
        act_name = body['actions'][0]['name']

        resp, body = self.client.get_definition('actions', act_name)
        self.assertEqual(200, resp.status)
        self.assertIsNotNone(body)
        self.assertIn(act_name, body)

    @decorators.attr(type='negative')
    @decorators.idempotent_id('c2b5be88-571a-4855-922f-9a338dba6adb')
    def test_get_nonexistent_action(self):
        self.assertRaises(
            exceptions.NotFound,
            self.client.get_object,
            'actions', 'nonexist'
        )

    @decorators.attr(type='negative')
    @decorators.idempotent_id('fc2fafcb-9bb4-4a18-a507-3f9964f4a08a')
    def test_double_creation(self):
        self.client.create_action('action_v2.yaml')

        self.assertRaises(
            exceptions.Conflict,
            self.client.create_action,
            'action_v2.yaml'
        )

    @decorators.attr(type='negative')
    @decorators.idempotent_id('0c456a73-9c39-4aeb-b3ca-3ea4338bc9ab')
    def test_create_action_invalid_def(self):
        self.assertRaises(
            exceptions.BadRequest,
            self.client.create_action,
            'wb_v2.yaml'
        )

    @decorators.attr(type='negative')
    @decorators.idempotent_id('469677b5-22ab-4e2a-aee6-5bcc9dac93de')
    def test_update_action_invalid_def(self):
        self.assertRaises(
            exceptions.BadRequest,
            self.client.update_request,
            'actions',
            'wb_v2.yaml'
        )

    @decorators.attr(type='negative')
    @decorators.idempotent_id('ab444607-40fc-47cb-982f-83762d5b64c9')
    def test_delete_nonexistent_action(self):
        self.assertRaises(
            exceptions.NotFound,
            self.client.delete_obj,
            'actions',
            'nonexist'
        )

    # TODO(rakhmerov): Rewrite it after https://review.opendev.org/#/c/746022/
    # is merged.
    @testtools.skip('Needs to be rewritten.')
    @decorators.attr(type='negative')
    @decorators.idempotent_id('74d0d480-793a-46ca-b88a-8336c1897f3a')
    def test_delete_standard_action(self):
        self.assertRaises(
            exceptions.BadRequest,
            self.client.delete_obj,
            'actions',
            'nova.servers_create'
        )



# Copyright 2013 Mirantis, Inc. All Rights Reserved.
# Copyright 2016 NEC Corporation. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import time

from oslo_serialization import jsonutils
from tempest import config
from tempest.lib import auth
from tempest.lib.common import rest_client
from tempest.lib import exceptions

CONF = config.CONF


def get_resource(path):
    main_package = 'mistral_tempest_tests'
    dir_path = __file__[0:__file__.find(main_package)]

    return open(dir_path +
                'mistral_tempest_tests/tests/resources/' +
                path).read()


def find_items(items, **props):
    def _matches(item, **props):
        for prop_name, prop_val in props.items():
            if item[prop_name] != prop_val:
                return False

        return True

    filtered = list([item for item in items if _matches(item, **props)])

    if len(filtered) == 1:
        return filtered[0]

    return filtered


class MistralClientBase(rest_client.RestClient):
    def __init__(self, auth_provider, service_type):
        super(MistralClientBase, self).__init__(
            auth_provider=auth_provider,
            service=service_type,
            region=CONF.identity.region,
            disable_ssl_certificate_validation=True
        )

        if service_type not in ('workflow', 'workflowv2'):
            msg = "Invalid parameter 'service_type'. "
            raise exceptions.UnprocessableEntity(msg)

        self.endpoint_url = 'publicURL'

        self.workbooks = []
        self.executions = []
        self.workflows = []
        self.triggers = []
        self.actions = []
        self.action_executions = []
        self.event_triggers = []

    def get_list_obj(self, url_path):
        resp, body = self.get(url_path)

        return resp, jsonutils.loads(body)

    def delete_obj(self, obj, name, force=None):
        if force:
            # TODO(apetrich) This try except is a partial implementation
            # to be removed once force deletion lands in mistral
            try:
                return self.delete('{obj}/{name}?force={force}'.format(
                    obj=obj,
                    name=name,
                    force=bool(force))
                )
            except Exception:
                pass

        return self.delete('{obj}/{name}'.format(obj=obj, name=name))

    def get_object(self, obj, _id):
        resp, body = self.get('{obj}/{id}'.format(obj=obj, id=_id))

        return resp, jsonutils.loads(body)

    def wait_execution_success(self, ex_body, timeout=180, url='executions'):
        return self.wait_execution(ex_body, timeout=timeout, url=url)

    def wait_execution(self, ex_body, timeout=180, url='executions',
                       target_state='SUCCESS'):
        start_time = time.time()

        expected_states = [target_state, 'RUNNING']

        while ex_body['state'] != target_state:
            if time.time() - start_time > timeout:
                msg = ("Execution exceeds timeout {0} "
                       "to change state to {1}. "
                       "Execution: {2}".format(timeout, target_state, ex_body))
                raise exceptions.TimeoutException(msg)

            _, ex_body = self.get_object(url, ex_body['id'])

            if ex_body['state'] not in expected_states:
                msg = ("Execution state %s is not in expected "
                       "states: %s" % (ex_body['state'], expected_states))
                raise exceptions.TempestException(msg)

            time.sleep(1)

        return ex_body


class AuthProv(auth.KeystoneV2AuthProvider):
    def __init__(self):
        self.alt_part = None

    def auth_request(self, method, url, *args, **kwargs):
        req_url, headers, body = super(AuthProv, self).auth_request(
            method, url, *args, **kwargs)
        return 'http://localhost:8989/{0}/{1}'.format(
            os.environ['VERSION'], url), headers, body

    def get_auth(self):
        return 'mock_str', 'mock_str'

    def base_url(self, *args, **kwargs):
        return ''



# Copyright 2016 NEC Corporation. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
from unittest import mock

from tempest import config
from tempest import test as test

from mistral_tempest_tests.services import base as service_base
from mistral_tempest_tests.services.v2 import mistral_client

CONF = config.CONF


class TestCase(test.BaseTestCase):
    credentials = ['admin', 'primary', 'alt']

    @classmethod
    def skip_checks(cls):
        super(TestCase, cls).skip_checks()

        if not CONF.service_available.mistral:
            raise cls.skipException("Mistral support is required.")

    def setUp(self):
        """Client authentication.

        This method allows to initialize authentication before
        each test case and define parameters of Mistral API Service.
        """
        super(TestCase, self).setUp()

        if 'WITHOUT_AUTH' in os.environ:
            self.mgr = mock.MagicMock()
            self.mgr.auth_provider = service_base.AuthProv()
            self.admin_mgr = self.alt_mgr = self.mgr
        else:
            self.admin_mgr = self.os_admin
            self.mgr = self.os_primary
            self.alt_mgr = self.os_alt

        if self._service == 'workflowv2':
            self.admin_client = mistral_client.MistralClientV2(
                self.admin_mgr.auth_provider, self._service)
            self.client = mistral_client.MistralClientV2(
                self.mgr.auth_provider, self._service)
            self.alt_client = mistral_client.MistralClientV2(
                self.alt_mgr.auth_provider, self._service)

    def assertPartIn(self, needle, haystack, message=''):
        self.assertTrue(any(needle in s for s in haystack), message)

    def tearDown(self):
        super(TestCase, self).tearDown()

        for wb in self.client.workbooks:
            self.client.delete_obj('workbooks', wb)

        self.client.workbooks = []


class TestCaseAdvanced(TestCase):

    def setUp(self):
        super(TestCaseAdvanced, self).setUp()

        self.image_ref = CONF.compute.image_ref
        self.flavor_ref = CONF.compute.flavor_ref

    def tearDown(self):
        for wb in self.client.workbooks:
            self.client.delete_obj('workbooks', wb)

        self.client.workbooks = []

        for ex in self.client.executions:
            self.client.delete_obj('executions', ex, force=True)

        self.client.executions = []

        super(TestCaseAdvanced, self).tearDown()



# Copyright 2013 - Mirantis, Inc.
# Copyright 2015 - Huawei Technologies Co. Ltd
# Copyright 2016 - Brocade Communications Systems, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import contextlib
import os
import shutil
import tempfile

from oslo_concurrency import processutils
from oslo_serialization import jsonutils


class NotDefined(object):
    """Marker of an empty value.

    In a number of cases None can't be used to express the semantics of
    a not defined value because None is just a normal value rather than
    a value set to denote that it's not defined. This class can be used
    in such cases instead of None.
    """

    pass


def get_dict_from_string(string, delimiter=','):
    if not string:
        return {}

    kv_dicts = []

    for kv_pair_str in string.split(delimiter):
        kv_str = kv_pair_str.strip()
        kv_list = kv_str.split('=')

        if len(kv_list) > 1:
            try:
                value = jsonutils.loads(kv_list[1])
            except ValueError:
                value = kv_list[1]

            kv_dicts += [{kv_list[0]: value}]
        else:
            kv_dicts += [kv_list[0]]

    return get_dict_from_entries(kv_dicts)


def get_dict_from_entries(entries):
    """Transforms a list of entries into dictionary.

    :param entries: A list of entries.
        If an entry is a dictionary the method simply updates the result
        dictionary with its content.
        If an entry is not a dict adds {entry, NotDefined} into the result.
    """

    result = {}

    for e in entries:
        if isinstance(e, dict):
            result.update(e)
        else:
            # NOTE(kong): we put NotDefined here as the value of
            # param without value specified, to distinguish from
            # the valid values such as None, ''(empty string), etc.
            result[e] = NotDefined

    return result


@contextlib.contextmanager
def tempdir(**kwargs):
    argdict = kwargs.copy()

    if 'dir' not in argdict:
        argdict['dir'] = '/tmp/'

    tmpdir = tempfile.mkdtemp(**argdict)

    try:
        yield tmpdir
    finally:
        try:
            shutil.rmtree(tmpdir)
        except OSError as e:
            raise OSError(
                "Failed to delete temp dir %(dir)s (reason: %(reason)s)" %
                {'dir': tmpdir, 'reason': e}
            )


def save_text_to(text, file_path, overwrite=False):
    if os.path.exists(file_path) and not overwrite:
        raise OSError(
            "Cannot save data to file. File %s already exists."
        )

    with open(file_path, 'w') as f:
        f.write(text)


def generate_key_pair(key_length=2048):
    """Create RSA key pair with specified number of bits in key.

    Returns tuple of private and public keys.
    """
    with tempdir() as tmpdir:
        keyfile = os.path.join(tmpdir, 'tempkey')
        args = [
            'ssh-keygen',
            '-q',  # quiet
            '-N', '',  # w/o passphrase
            '-t', 'rsa',  # create key of rsa type
            '-f', keyfile,  # filename of the key file
            '-C', 'Generated-by-Mistral'  # key comment
        ]

        if key_length is not None:
            args.extend(['-b', key_length])

        processutils.execute(*args)

        if not os.path.exists(keyfile):
            # raise exc.DataAccessException(
            #     "Private key file hasn't been created"
            # )
            raise OSError("Private key file hasn't been created")

        private_key = open(keyfile).read()
        public_key_path = keyfile + '.pub'

        if not os.path.exists(public_key_path):
            # raise exc.DataAccessException(
            #     "Public key file hasn't been created"
            # )
            raise OSError("Private key file hasn't been created")
        public_key = open(public_key_path).read()

        return private_key, public_key



# Copyright 2016 NEC Corporation. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_serialization import jsonutils
from oslo_utils import uuidutils
from tempest import config

from mistral_tempest_tests.services import base

CONF = config.CONF


class MistralClientV2(base.MistralClientBase):

    def post_request(self, url_path, file_name):
        headers = {"headers": "Content-Type:text/plain"}

        return self.post(
            url_path,
            base.get_resource(file_name),
            headers=headers
        )

    def get_request(self, url_path):
        headers = {"headers": "Content-Type:application/json"}

        return self.get(url_path, headers=headers)

    def post_json(self, url_path, obj, extra_headers=None):
        if extra_headers is None:
            extra_headers = {}
        headers = {"Content-Type": "application/json"}
        headers = dict(headers, **extra_headers)
        return self.post(url_path,
                         jsonutils.dump_as_bytes(obj),
                         headers=headers)

    def update_request(self, url_path, file_name):
        headers = {"headers": "Content-Type:text/plain"}

        resp, body = self.put(
            url_path,
            base.get_resource(file_name),
            headers=headers
        )

        return resp, jsonutils.loads(body)

    def get_definition(self, item, name):
        resp, body = self.get("%s/%s" % (item, name))

        return resp, jsonutils.loads(body)['definition']

    def create_workbook(self, yaml_file):
        resp, body = self.post_request('workbooks', yaml_file)

        wb_name = jsonutils.loads(body)['name']
        self.workbooks.append(wb_name)

        _, wfs = self.get_list_obj('workflows')

        for wf in wfs['workflows']:
            if wf['name'].startswith(wb_name):
                self.workflows.append(wf['id'])

        return resp, jsonutils.loads(body)

    def create_workflow(self, yaml_file, scope=None, namespace=None):
        url_path = 'workflows?'

        if scope:
            url_path += 'scope=public&'

        if namespace:
            url_path += 'namespace=' + namespace

        resp, body = self.post_request(url_path, yaml_file)

        for wf in jsonutils.loads(body)['workflows']:
            self.workflows.append(wf['id'])

        return resp, jsonutils.loads(body)

    def get_workflow(self, wf_identifier, namespace=None):

        url_path = 'workflows/' + wf_identifier
        if namespace:
            url_path += 'namespace=' + namespace

        resp, body = self.get_request(url_path)

        return resp, jsonutils.loads(body)

    def update_workflow(self, file_name, namespace=None):
        url_path = "workflows?"

        if namespace:
            url_path += 'namespace=' + namespace

        return self.update_request(url_path, file_name=file_name)

    def get_action_execution(self, action_execution_id):
        return self.get('action_executions/%s' % action_execution_id)

    def get_action_executions(self, task_id=None):
        url_path = 'action_executions'
        if task_id:
            url_path += '?task_execution_id=%s' % task_id

        return self.get_list_obj(url_path)

    def create_execution(self, identifier, wf_namespace=None, wf_input=None,
                         params=None):
        if uuidutils.is_uuid_like(identifier):
            body = {"workflow_id": "%s" % identifier}
        else:
            body = {"workflow_name": "%s" % identifier}

        if wf_namespace:
            body.update({'workflow_namespace': wf_namespace})

        if wf_input:
            body.update({'input': jsonutils.dump_as_bytes(wf_input)})
        if params:
            body.update({'params': jsonutils.dump_as_bytes(params)})

        resp, body = self.post('executions', jsonutils.dump_as_bytes(body))

        self.executions.append(jsonutils.loads(body)['id'])

        return resp, jsonutils.loads(body)

    def update_execution(self, execution_id, put_body):
        resp, body = self.put('executions/%s' % execution_id, put_body)

        return resp, jsonutils.loads(body)

    def get_execution(self, execution_id):
        return self.get('executions/%s' % execution_id)

    def get_executions(self, task_id):
        url_path = 'executions'
        if task_id:
            url_path += '?task_execution_id=%s' % task_id

        return self.get_list_obj(url_path)

    def get_tasks(self, execution_id=None):
        url_path = 'tasks'
        if execution_id:
            url_path += '?workflow_execution_id=%s' % execution_id

        return self.get_list_obj(url_path)

    def create_cron_trigger(self, name, wf_name, wf_input=None, pattern=None,
                            first_time=None, count=None):
        post_body = {
            'name': name,
            'workflow_name': wf_name,
            'pattern': pattern,
            'remaining_executions': count,
            'first_execution_time': first_time
        }

        if wf_input:
            post_body.update({
                'workflow_input': jsonutils.dump_as_bytes(wf_input)})

        rest, body = self.post('cron_triggers',
                               jsonutils.dump_as_bytes(post_body))

        self.triggers.append(name)

        return rest, jsonutils.loads(body)

    def create_action(self, yaml_file):
        resp, body = self.post_request('actions', yaml_file)

        self.actions.extend(
            [action['name'] for action in jsonutils.loads(body)['actions']])

        return resp, jsonutils.loads(body)

    def get_wf_tasks(self, wf_name):
        all_tasks = self.get_list_obj('tasks')[1]['tasks']

        return [t for t in all_tasks if t['workflow_name'] == wf_name]

    def create_action_execution(self, request_body, extra_headers=None):
        if extra_headers is None:
            extra_headers = {}
        resp, body = self.post_json('action_executions', request_body,
                                    extra_headers)

        params = jsonutils.loads(request_body.get('params', '{}'))
        if params.get('save_result', False):
            self.action_executions.append(jsonutils.loads(body)['id'])

        return resp, jsonutils.loads(body)

    def create_event_trigger(self, wf_id, exchange, topic, event, name='',
                             wf_input=None, wf_params=None):
        post_body = {
            'workflow_id': wf_id,
            'exchange': exchange,
            'topic': topic,
            'event': event,
            'name': name
        }

        if wf_input:
            post_body.update({
                'workflow_input': jsonutils.dump_as_bytes(wf_input)})

        if wf_params:
            post_body.update({
                'workflow_params': jsonutils.dump_as_bytes(wf_params)})

        rest, body = self.post('event_triggers',
                               jsonutils.dump_as_bytes(post_body))

        event_trigger = jsonutils.loads(body)
        self.event_triggers.append(event_trigger['id'])

        return rest, event_trigger
